/*
Copyright (c) 2019-2022 Digital Energy Cloud Solutions LLC. All Rights Reserved.
Authors:
Petr Krutov, <petr.krutov@digitalenergy.online>
Stanislav Solovev, <spsolovev@digitalenergy.online>

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

/*
Terraform DECORT provider - manage resources provided by DECORT (Digital Energy Cloud
Orchestration Technology) with Terraform by Hashicorp.

Source code: https://repository.basistech.ru/BASIS/terraform-provider-decort

Please see README.md to learn where to place source code so that it
builds seamlessly.

Documentation: https://repository.basistech.ru/BASIS/terraform-provider-decort/wiki
*/

package account

import (
	"context"
	"strconv"
	"strings"

	"github.com/hashicorp/terraform-plugin-sdk/v2/diag"
	"github.com/hashicorp/terraform-plugin-sdk/v2/helper/schema"
	log "github.com/sirupsen/logrus"
	"repository.basistech.ru/BASIS/decort-golang-sdk/pkg/cloudbroker/account"
	"repository.basistech.ru/BASIS/terraform-provider-decort/internal/constants"
	"repository.basistech.ru/BASIS/terraform-provider-decort/internal/controller"
)

func resourceAccountCreate(ctx context.Context, d *schema.ResourceData, m interface{}) diag.Diagnostics {
	log.Debugf("resourceAccountCreate")

	c := m.(*controller.ControllerCfg)
	req := account.CreateRequest{}

	req.Name = d.Get("account_name").(string)
	req.Username = d.Get("username").(string)

	if emailaddress, ok := d.GetOk("emailaddress"); ok {
		req.EmailAddress = emailaddress.(string)
	}
	if sendAccessEmails, ok := d.GetOk("send_access_emails"); ok {
		req.SendAccessEmails = sendAccessEmails.(bool)
	}
	if resLimits, ok := d.GetOk("resource_limits"); ok {
		resLimit := resLimits.([]interface{})[0]
		resLimitConv := resLimit.(map[string]interface{})
		if resLimitConv["cu_m"] != nil {
			maxMemCap := int64(resLimitConv["cu_m"].(float64))
			if maxMemCap == 0 {
				req.MaxMemoryCapacity = -1
			} else {
				req.MaxMemoryCapacity = maxMemCap
			}
		}
		if resLimitConv["cu_d"] != nil {
			maxDiskCap := int64(resLimitConv["cu_d"].(float64))
			if maxDiskCap == 0 {
				req.MaxVDiskCapacity = -1
			} else {
				req.MaxVDiskCapacity = maxDiskCap
			}
		}
		if resLimitConv["cu_c"] != nil {
			maxCPUCap := int64(resLimitConv["cu_c"].(float64))
			if maxCPUCap == 0 {
				req.MaxCPUCapacity = -1
			} else {
				req.MaxCPUCapacity = maxCPUCap
			}

		}
		if resLimitConv["cu_i"] != nil {
			maxNumPublicIP := int64(resLimitConv["cu_i"].(float64))
			if maxNumPublicIP == 0 {
				req.MaxNumPublicIP = -1
			} else {
				req.MaxNumPublicIP = maxNumPublicIP
			}

		}
		if resLimitConv["cu_np"] != nil {
			maxNP := int64(resLimitConv["cu_np"].(float64))
			if maxNP == 0 {
				req.MaxNetworkPeerTransfer = -1
			} else {
				req.MaxNetworkPeerTransfer = maxNP
			}

		}
		if resLimitConv["gpu_units"] != nil {
			gpuUnits := int64(resLimitConv["gpu_units"].(float64))
			if gpuUnits == 0 {
				req.GPUUnits = -1
			} else {
				req.GPUUnits = gpuUnits
			}
		}
	}

	accountId, err := c.CloudBroker().Account().Create(ctx, req)
	if err != nil {
		return diag.FromErr(err)
	}

	d.SetId(strconv.FormatUint(accountId, 10))
	d.Set("account_id", accountId)

	diagnostics := resourceAccountRead(ctx, d, m)
	if diagnostics != nil {
		return diagnostics
	}

	if enable, ok := d.GetOk("enable"); ok {
		enable := enable.(bool)
		if enable {
			req := account.EnableRequest{
				AccountID: accountId,
			}
			_, err := c.CloudBroker().Account().Enable(ctx, req)
			if err != nil {
				return diag.FromErr(err)
			}
		} else {
			req := account.DisableRequest{
				AccountID: accountId,
			}
			_, err := c.CloudBroker().Account().Disable(ctx, req)
			if err != nil {
				return diag.FromErr(err)
			}
		}
	}

	if users, ok := d.GetOk("users"); ok {
		addedUsers := users.([]interface{})

		if len(addedUsers) > 0 {
			for _, user := range addedUsers {
				userConv := user.(map[string]interface{})
				req := account.AddUserRequest{
					AccountID:  accountId,
					Username:   userConv["user_id"].(string),
					AccessType: userConv["access_type"].(string),
				}

				_, err := c.CloudBroker().Account().AddUser(ctx, req)
				if err != nil {
					return diag.FromErr(err)
				}
			}
		}
	}

	return nil
}

func resourceAccountRead(ctx context.Context, d *schema.ResourceData, m interface{}) diag.Diagnostics {
	log.Debugf("resourceAccountRead")

	acc, err := utilityAccountCheckPresence(ctx, d, m)
	if err != nil {
		d.SetId("")
		return diag.FromErr(err)
	}

	flattenResourceAccount(d, acc)

	return nil
}

func resourceAccountDelete(ctx context.Context, d *schema.ResourceData, m interface{}) diag.Diagnostics {
	log.Debugf("resourceAccountDelete")

	accountData, err := utilityAccountCheckPresence(ctx, d, m)
	if err != nil {
		return diag.FromErr(err)
	}

	c := m.(*controller.ControllerCfg)
	req := account.DeleteRequest{
		AccountID:   accountData.ID,
		Permanently: d.Get("permanently").(bool),
	}

	_, err = c.CloudBroker().Account().Delete(ctx, req)
	if err != nil {
		return diag.FromErr(err)
	}

	d.SetId("")

	return nil
}

func resourceAccountUpdate(ctx context.Context, d *schema.ResourceData, m interface{}) diag.Diagnostics {
	log.Debugf("resourceAccountUpdate")
	c := m.(*controller.ControllerCfg)

	acc, err := utilityAccountCheckPresence(ctx, d, m)
	if err != nil {
		d.SetId("")
		return diag.FromErr(err)
	}

	if d.HasChange("enable") {
		enable := d.Get("enable").(bool)
		if enable {
			req := account.EnableRequest{
				AccountID: acc.ID,
			}
			_, err := c.CloudBroker().Account().Enable(ctx, req)
			if err != nil {
				return diag.FromErr(err)
			}
		} else {
			req := account.DisableRequest{
				AccountID: acc.ID,
			}
			_, err := c.CloudBroker().Account().Disable(ctx, req)
			if err != nil {
				return diag.FromErr(err)
			}
		}
	}

	updateReq := account.UpdateRequest{AccountID: acc.ID}
	hasChange := false

	if d.HasChange("account_name") {
		updateReq.Name = d.Get("account_name").(string)

		hasChange = true
	}
	if d.HasChange("resource_limits") {
		resLimit := d.Get("resource_limits").([]interface{})[0]
		resLimitConv := resLimit.(map[string]interface{})

		if resLimitConv["cu_m"] != nil {
			maxMemCap := int(resLimitConv["cu_m"].(float64))
			if maxMemCap == 0 {
				updateReq.MaxMemoryCapacity = -1
			} else {
				updateReq.MaxMemoryCapacity = int64(maxMemCap)
			}
		}
		if resLimitConv["cu_d"] != nil {
			maxDiskCap := int(resLimitConv["cu_d"].(float64))
			if maxDiskCap == 0 {
				updateReq.MaxVDiskCapacity = -1
			} else {
				updateReq.MaxVDiskCapacity = int64(maxDiskCap)
			}
		}
		if resLimitConv["cu_c"] != nil {
			maxCPUCap := int(resLimitConv["cu_c"].(float64))
			if maxCPUCap == 0 {
				updateReq.MaxCPUCapacity = -1
			} else {
				updateReq.MaxCPUCapacity = int64(maxCPUCap)
			}

		}
		if resLimitConv["cu_i"] != nil {
			maxNumPublicIP := int(resLimitConv["cu_i"].(float64))
			if maxNumPublicIP == 0 {
				updateReq.MaxNumPublicIP = -1
			} else {
				updateReq.MaxNumPublicIP = int64(maxNumPublicIP)
			}

		}
		if resLimitConv["cu_np"] != nil {
			maxNP := int(resLimitConv["cu_np"].(float64))
			if maxNP == 0 {
				updateReq.MaxNetworkPeerTransfer = -1
			} else {
				updateReq.MaxNetworkPeerTransfer = int64(maxNP)
			}

		}
		if resLimitConv["gpu_units"] != nil {
			gpuUnits := int(resLimitConv["gpu_units"].(float64))
			if gpuUnits == 0 {
				updateReq.GPUUnits = -1
			} else {
				updateReq.GPUUnits = int64(gpuUnits)
			}
		}

		hasChange = true
	}

	if d.HasChange("send_access_emails") {
		updateReq.SendAccessEmails = d.Get("send_access_emails").(bool)
		hasChange = true
	}

	if hasChange {
		_, err := c.CloudBroker().Account().Update(ctx, updateReq)
		if err != nil {
			return diag.FromErr(err)
		}
	}

	if d.HasChange("restore") {
		restore := d.Get("restore").(bool)
		if restore {
			req := account.RestoreRequest{
				AccountID: acc.ID,
			}

			_, err := c.CloudBroker().Account().Restore(ctx, req)
			if err != nil {
				return diag.FromErr(err)
			}
		}
	}

	if d.HasChange("users") {
		deletedUsers := make([]interface{}, 0)
		addedUsers := make([]interface{}, 0)
		updatedUsers := make([]interface{}, 0)

		old, new := d.GetChange("users")
		oldConv := old.([]interface{})
		newConv := new.([]interface{})
		for _, el := range oldConv {
			if !isContainsUser(newConv, el) {
				deletedUsers = append(deletedUsers, el)
			}
		}
		for _, el := range newConv {
			if !isContainsUser(oldConv, el) {
				addedUsers = append(addedUsers, el)
			} else {
				if isChangedUser(oldConv, el) {
					updatedUsers = append(updatedUsers, el)
				}
			}
		}

		if len(deletedUsers) > 0 {
			for _, user := range deletedUsers {
				userConv := user.(map[string]interface{})
				req := account.DeleteUserRequest{
					AccountID:       acc.ID,
					UserName:        userConv["user_id"].(string),
					RecursiveDelete: userConv["recursive_delete"].(bool),
				}

				_, err := c.CloudBroker().Account().DeleteUser(ctx, req)
				if err != nil {
					return diag.FromErr(err)
				}
			}
		}

		if len(addedUsers) > 0 {
			for _, user := range addedUsers {
				userConv := user.(map[string]interface{})
				req := account.AddUserRequest{
					AccountID:  acc.ID,
					Username:   userConv["user_id"].(string),
					AccessType: strings.ToUpper(userConv["access_type"].(string)),
				}

				_, err := c.CloudBroker().Account().AddUser(ctx, req)
				if err != nil {
					return diag.FromErr(err)
				}
			}
		}

		if len(updatedUsers) > 0 {
			for _, user := range updatedUsers {
				userConv := user.(map[string]interface{})
				req := account.UpdateUserRequest{
					AccountID:  acc.ID,
					UserID:     userConv["user_id"].(string),
					AccessType: strings.ToUpper(userConv["access_type"].(string)),
				}

				_, err := c.CloudBroker().Account().UpdateUser(ctx, req)
				if err != nil {
					return diag.FromErr(err)
				}
			}
		}

	}

	return nil
}

func isContainsUser(els []interface{}, el interface{}) bool {
	for _, elOld := range els {
		elOldConv := elOld.(map[string]interface{})
		elConv := el.(map[string]interface{})
		if elOldConv["user_id"].(string) == elConv["user_id"].(string) {
			return true
		}
	}
	return false
}

func isChangedUser(els []interface{}, el interface{}) bool {
	for _, elOld := range els {
		elOldConv := elOld.(map[string]interface{})
		elConv := el.(map[string]interface{})
		if elOldConv["user_id"].(string) == elConv["user_id"].(string) &&
			(!strings.EqualFold(elOldConv["access_type"].(string), elConv["access_type"].(string)) ||
				elOldConv["recursive_delete"].(bool) != elConv["recursive_delete"].(bool)) {
			return true
		}
	}
	return false
}

func ResourceAccount() *schema.Resource {
	return &schema.Resource{
		SchemaVersion: 1,

		CreateContext: resourceAccountCreate,
		ReadContext:   resourceAccountRead,
		UpdateContext: resourceAccountUpdate,
		DeleteContext: resourceAccountDelete,

		Importer: &schema.ResourceImporter{
			StateContext: schema.ImportStatePassthroughContext,
		},

		Timeouts: &schema.ResourceTimeout{
			Create:  &constants.Timeout60s,
			Read:    &constants.Timeout30s,
			Update:  &constants.Timeout60s,
			Delete:  &constants.Timeout60s,
			Default: &constants.Timeout60s,
		},

		Schema: resourceAccountSchemaMake(),
	}
}
